from collections.abc import Callable, Mapping, Sequence, Set
from datetime import date, datetime, time, timedelta, timezone
from enum import Enum
from pathlib import Path
from re import Pattern
from types import MappingProxyType, NoneType, UnionType
from typing import Any, Literal, Protocol, Union
from typing import Mapping as MappingType  # noqa: UP035
from typing import Sequence as SequenceType  # noqa: UP035
from typing import Sequence as SetType  # noqa: UP035
from uuid import UUID

from haiway.state.attributes import AttributeAnnotation
from haiway.types import MISSING, Missing

__all__ = [
    "attribute_validator",
]


def attribute_validator(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    if validator := VALIDATORS.get(annotation.origin):
        return validator(annotation)

    elif hasattr(annotation.origin, "__IMMUTABLE__"):
        return _prepare_validator_of_type(annotation)

    elif issubclass(annotation.origin, Protocol):
        return _prepare_validator_of_type(annotation)

    elif issubclass(annotation.origin, Enum):
        return _prepare_validator_of_type(annotation)

    else:
        raise TypeError(f"Unsupported type annotation: {annotation}")


def _prepare_validator_of_any(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    def validator(
        value: Any,
    ) -> Any:
        return value  # any is always valid

    return validator


def _prepare_validator_of_none(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    def validator(
        value: Any,
    ) -> Any:
        if value is None:
            return value

        else:
            raise TypeError(f"'{value}' is not matching expected type of 'None'")

    return validator


def _prepare_validator_of_missing(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    def validator(
        value: Any,
    ) -> Any:
        if value is MISSING:
            return value

        else:
            raise TypeError(f"'{value}' is not matching expected type of 'Missing'")

    return validator


def _prepare_validator_of_literal(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    elements: list[Any] = annotation.arguments
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
    ) -> Any:
        if value in elements:
            return value

        else:
            raise ValueError(f"'{value}' is not matching expected values of '{formatted_type}'")

    return validator


def _prepare_validator_of_type(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    validated_type: type[Any] = annotation.origin
    formatted_type: str = str(annotation)

    def type_validator(
        value: Any,
    ) -> Any:
        match value:
            case value if isinstance(value, validated_type):
                return value

            case _:
                raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return type_validator


def _prepare_validator_of_set(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    element_validator: Callable[[Any], Any] = attribute_validator(annotation.arguments[0])
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
    ) -> Any:
        if isinstance(value, set):
            return frozenset(element_validator(element) for element in value)  # pyright: ignore[reportUnknownVariableType]

        else:
            raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_sequence(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    element_validator: Callable[[Any], Any] = attribute_validator(annotation.arguments[0])
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
    ) -> Any:
        match value:
            case [*elements]:
                return tuple(element_validator(element) for element in elements)

            case _:
                raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_mapping(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    key_validator: Callable[[Any], Any] = attribute_validator(annotation.arguments[0])
    value_validator: Callable[[Any], Any] = attribute_validator(annotation.arguments[1])
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
    ) -> Any:
        match value:
            case {**elements}:
                return MappingProxyType(
                    {key_validator(key): value_validator(value) for key, value in elements.items()}
                )

            case _:
                raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_tuple(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    if annotation.arguments[-1].origin == Ellipsis:
        element_validator: Callable[[Any], Any] = attribute_validator(annotation.arguments[0])
        formatted_type: str = str(annotation)

        def validator(
            value: Any,
        ) -> Any:
            match value:
                case [*elements]:
                    return tuple(element_validator(element) for element in elements)

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

        return validator

    else:
        element_validators: list[Callable[[Any], Any]] = [
            attribute_validator(alternative) for alternative in annotation.arguments
        ]
        elements_count: int = len(element_validators)
        formatted_type: str = str(annotation)

        def validator(
            value: Any,
        ) -> Any:
            match value:
                case [*elements]:
                    if len(elements) != elements_count:
                        raise ValueError(
                            f"'{value}' is not matching expected type of '{formatted_type}'"
                        )

                    return tuple(
                        element_validators[idx](value) for idx, value in enumerate(elements)
                    )

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

        return validator


def _prepare_validator_of_union(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    validators: list[Callable[[Any], Any]] = [
        attribute_validator(alternative) for alternative in annotation.arguments
    ]
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
    ) -> Any:
        errors: list[Exception] = []
        for validator in validators:
            try:
                return validator(value)

            except Exception as exc:
                errors.append(exc)

        raise ExceptionGroup(
            f"'{value}' is not matching expected type of '{formatted_type}'",
            errors,
        )

    return validator


def _prepare_validator_of_callable(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
    ) -> Any:
        if callable(value):
            return value

        else:
            raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


VALIDATORS: Mapping[Any, Callable[[AttributeAnnotation], Callable[[Any], Any]]] = {
    Any: _prepare_validator_of_any,
    NoneType: _prepare_validator_of_none,
    Missing: _prepare_validator_of_missing,
    type: _prepare_validator_of_type,
    bool: _prepare_validator_of_type,
    int: _prepare_validator_of_type,
    float: _prepare_validator_of_type,
    complex: _prepare_validator_of_type,
    bytes: _prepare_validator_of_type,
    str: _prepare_validator_of_type,
    tuple: _prepare_validator_of_tuple,
    frozenset: _prepare_validator_of_set,
    Literal: _prepare_validator_of_literal,
    set: _prepare_validator_of_set,
    Set: _prepare_validator_of_set,
    SetType: _prepare_validator_of_set,
    Sequence: _prepare_validator_of_sequence,
    SequenceType: _prepare_validator_of_sequence,
    Mapping: _prepare_validator_of_mapping,
    MappingType: _prepare_validator_of_mapping,
    range: _prepare_validator_of_type,
    UUID: _prepare_validator_of_type,
    date: _prepare_validator_of_type,
    datetime: _prepare_validator_of_type,
    time: _prepare_validator_of_type,
    timedelta: _prepare_validator_of_type,
    timezone: _prepare_validator_of_type,
    Path: _prepare_validator_of_type,
    Pattern: _prepare_validator_of_type,
    Union: _prepare_validator_of_union,
    UnionType: _prepare_validator_of_union,
    Callable: _prepare_validator_of_callable,
}



from haiway.helpers.asynchrony import asynchronous, wrap_async
from haiway.helpers.caching import cache
from haiway.helpers.retries import retry
from haiway.helpers.throttling import throttle
from haiway.helpers.timeouted import timeout
from haiway.helpers.tracing import ArgumentsTrace, ResultTrace, traced

__all__ = [
    "ArgumentsTrace",
    "ResultTrace",
    "asynchronous",
    "cache",
    "retry",
    "throttle",
    "timeout",
    "traced",
    "wrap_async",
]



from collections.abc import Callable, Sequence, Set
from datetime import date, datetime
from enum import StrEnum
from typing import Literal, Protocol, Self, runtime_checkable
from uuid import UUID, uuid4

from haiway import MISSING, Missing, State, frozenlist


def test_basic_initializes_with_arguments() -> None:
    class Selection(StrEnum):
        A = "A"
        B = "A"

    @runtime_checkable
    class Proto(Protocol):
        def __call__(self) -> None: ...

    class Basics(State):
        uuid: UUID
        date: date
        datetime: datetime
        string: str
        literal: Literal["A", "B"]
        sequence: Sequence[str]
        string_set: Set[str]
        frozen: frozenlist[int]
        integer: int
        union: str | int
        optional: str | None
        none: None
        function: Callable[[], None]
        proto: Proto
        selection: Selection

    basic = Basics(
        uuid=uuid4(),
        date=date.today(),
        datetime=datetime.now(),
        string="string",
        literal="A",
        sequence=["a", "b", "c"],
        string_set={"a", "b"},
        frozen=(1, 2, 3),
        integer=0,
        union="union",
        optional="optional",
        none=None,
        function=lambda: None,
        proto=lambda: None,
        selection=Selection.A,
    )
    assert basic.string == "string"
    assert basic.literal == "A"
    assert basic.sequence == ("a", "b", "c")
    assert basic.string_set == {"a", "b"}
    assert basic.frozen == (1, 2, 3)
    assert basic.integer == 0
    assert basic.union == "union"
    assert basic.optional == "optional"
    assert basic.none is None
    assert basic.selection == Selection.A
    assert callable(basic.function)
    assert isinstance(basic.proto, Proto)


def test_basic_initializes_with_defaults() -> None:
    class Basics(State):
        string: str = ""
        integer: int = 0
        optional: str | None = None

    basic = Basics()
    assert basic.string == ""
    assert basic.integer == 0
    assert basic.optional is None


def test_basic_equals_checks_properties() -> None:
    class Basics(State):
        string: str
        integer: int

    assert Basics(string="a", integer=1) == Basics(string="a", integer=1)
    assert Basics(string="a", integer=1) != Basics(string="b", integer=1)
    assert Basics(string="a", integer=1) != Basics(string="a", integer=2)
    assert Basics(string="a", integer=1) != Basics(string="b", integer=2)


def test_basic_initializes_with_arguments_and_defaults() -> None:
    class Basics(State):
        string: str
        integer: int = 0
        optional: str | None = None

    basic = Basics(
        string="string",
        integer=42,
    )
    assert basic.string == "string"
    assert basic.integer == 42
    assert basic.optional is None


def test_parametrized_initializes_with_proper_parameters() -> None:
    class Parametrized[T](State):
        value: T

    parametrized_string = Parametrized(
        value="string",
    )
    assert parametrized_string.value == "string"

    parametrized_int = Parametrized(
        value=42,
    )
    assert parametrized_int.value == 42

    assert parametrized_string != parametrized_int


def test_nested_initializes_with_proper_arguments() -> None:
    class Nested(State):
        string: str

    class Recursive(State):
        nested: Nested
        recursion: "Recursive | None"
        self_recursion: Self | None

    recursive = Recursive(
        nested=Nested(string="one"),
        recursion=Recursive(
            nested=Nested(string="two"),
            recursion=None,
            self_recursion=None,
        ),
        self_recursion=None,
    )
    assert recursive.nested == Nested(string="one")
    assert recursive.recursion == Recursive(
        nested=Nested(string="two"),
        recursion=None,
        self_recursion=None,
    )


def test_dict_skips_missing_properties() -> None:
    class Basics(State):
        string: str
        integer: int | Missing | None

    assert Basics(string="a", integer=1).as_dict() == {"string": "a", "integer": 1}
    assert Basics(string="a", integer=MISSING).as_dict() == {"string": "a"}
    assert Basics(string="a", integer=None).as_dict() == {"string": "a", "integer": None}


def test_initialization_allows_missing_properties() -> None:
    class Basics(State):
        string: str
        integer: int | Missing | None

    assert Basics(**{"string": "a", "integer": 1}) == Basics(string="a", integer=1)
    assert Basics(**{"string": "a", "integer": None}) == Basics(string="a", integer=None)
    assert Basics(**{"string": "a"}) == Basics(string="a", integer=MISSING)



from asyncio import gather
from collections.abc import Iterable
from contextlib import AbstractAsyncContextManager
from itertools import chain
from types import TracebackType
from typing import final

from haiway.state import State
from haiway.utils import freeze

__all__ = [
    "Disposable",
    "Disposables",
]

type Disposable = AbstractAsyncContextManager[Iterable[State] | State | None]


@final
class Disposables:
    def __init__(
        self,
        *disposables: Disposable,
    ) -> None:
        self._disposables: tuple[Disposable, ...] = disposables

        freeze(self)

    def __bool__(self) -> bool:
        return len(self._disposables) > 0

    async def _initialize(
        self,
        disposable: Disposable,
        /,
    ) -> Iterable[State]:
        match await disposable.__aenter__():
            case None:
                return ()

            case State() as single:
                return (single,)

            case multiple:
                return multiple

    async def __aenter__(self) -> Iterable[State]:
        return [
            *chain.from_iterable(
                state
                for state in await gather(
                    *[self._initialize(disposable) for disposable in self._disposables],
                )
            )
        ]

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        results: list[bool | BaseException | None] = await gather(
            *[
                disposable.__aexit__(
                    exc_type,
                    exc_val,
                    exc_tb,
                )
                for disposable in self._disposables
            ],
            return_exceptions=True,
        )

        exceptions: list[BaseException] = [exc for exc in results if isinstance(exc, BaseException)]

        if len(exceptions) > 1:
            raise BaseExceptionGroup("Disposing errors", exceptions)
